# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""
   Build a list of testcases using a list of command/keywords and/or
   verifying some criterias about cputime or memory.
   
   Supported parameters :  ['mem_aster', 'ncpus', 'tps_job', 'mem_job']
"""

__revision__ = "$Id: as_list.py 3309 2008-06-05 08:49:38Z courtois $"


# ----- differ messages translation
def _(mesg):
   return mesg

import sys
import os
import re
from sets         import Set
from glob         import glob

from as_config    import ASTER_CONFIG
from as_utils     import getpara, get_list
from progress     import Progress


#------------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class FILTER(object):
   """Permet de savoir si on conserve un test ou non.
   La mthode 'check' retourne True si on doit conserver le test, False sinon.
   Les arguments sont :
      test : nom du cas-test avec le rpertoire, sans l'extension
      para : dictionnaire des paramtres (fichier .para)
      jdc  : texte du jeu de commandes (fichiers *.com?)
   """

# -----------------------------------------------------------------------------
class FILTER_REGEXP(FILTER):
   """Le contenu du fichier de commandes doit vrifie une expression rgulire.
   """
   def __init__(self, s_regexp, flags=re.MULTILINE | re.DOTALL):
      """s_regexp : chaine de l'expression qui doit tre vrifie.
      """
      print " - regular expression : %s" % repr(s_regexp)
      self.regexp = re.compile(s_regexp, flags)

   def check(self, **kwargs):
      """Applique le filtre.
      Argument : jdc
      """
      content = kwargs.get('jdc', '')
      mat = self.regexp.search(content)
      return mat is not None

# -----------------------------------------------------------------------------
class FILTER_PARA(FILTER):
   """Les paramtres du cas-test doivent vrifie une condition.
   """
   def __init__(self, s_cond):
      """s_cond : texte de la condition
      """
      self.condition = s_cond

   def check(self, **kwargs):
      """Applique le filtre.
      Argument : para
      """
      para = kwargs.get('para', {})
      try:
         keep = eval(self.condition, {'d' : para})
      except KeyError:
         keep = False
      return keep

#------------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class LISTE_CT(object):
   """Classe permettant de lire un ensemble de fichiers de tests et d'effectuer
   un filtre selon certains critres afin d'extraire une liste de cas-test.
   """
   def __init__(self, astest_dir, all=False, verbose=True):
      """Initialisations
      """
      self.astest_dir = astest_dir
      self.liste_ct = []
      self.alltest = all
      self.filter = []
      self.liste_crit = []
      self.verbose = verbose
      if self.verbose:
         print 'Rpertoire des cas-tests : %s' % self.astest_dir

# -----------------------------------------------------------------------------
   def add_ct(self, l_file):
      """Ajoute des cas-tests dans la liste.
      """
      new_ct = []
      if len(l_file) == 0 and self.alltest:
         if self.verbose:
            print "Recherche de '%s'..." % self.filename('*', 'comm')
         l_file = '*'
      # sans extension
      for ct in l_file:
         l_gl = glob(self.filename(ct, 'comm'))
         new_ct.extend([os.path.basename(os.path.splitext(t)[0]) for t in l_gl])
      # tests existants
      new_ct = Set([ct for ct in new_ct if os.path.isfile(self.filename(ct, 'comm'))])
      new_ct.update(self.liste_ct)
      self.liste_ct = list(new_ct)
      self.liste_ct.sort()
      if self.verbose:
         print '%6d tests dans la liste.' % len(self.liste_ct)
   
# -----------------------------------------------------------------------------
   def filename(self, ct, ext):
      """Retourne le nom du fichier du cas-test.
      """
      return os.path.join(self.astest_dir, '%s.%s' % (ct, ext))
   
# -----------------------------------------------------------------------------
   def read_para_ct(self, ct):
      """Lecture du fichier .para d'un cas-test
      """
      return getpara(self.filename(ct, 'para'))

# -----------------------------------------------------------------------------
   def decode_critere(self, txt):
      """Rend le critre valuable sur le dictonnaire des paramtres d'un cas-test
      """
      mat = re.search(' *([a-z_]+) *([<>=])+ *([a-zA-Z0-9_]+) *', txt)
      crit = None
      if mat is None:
         print '>>> Critre invalide : %s' % txt
      else:
         key, comp, val = mat.groups()
         if key not in ['mem_aster', 'ncpus', 'tps_job', 'mem_job']:
            print '>>> Paramtre inconnu : %s' % key
         else:
            crit = "d['%s'] %s %s" % (key, comp, val)
      return crit

# -----------------------------------------------------------------------------
   def decode_cmde(self, txt0):
      """Retourne une expression rgulire pour filtrer sur les commandes
      utilises.
         txt = COMMANDE[/MOTCLEFACT[/MOTCLE[=VALEUR]]]
      """
      #txt = re.split('[/=]+', txt.strip())
      txt = txt0.strip().split('/')
      d = {
         'cmde' : txt.pop(0),
         'mcfact' : '',
         'mcsimp' : '',
         'valeur' : '',
      }
      if len(txt) > 0:
         dern = txt.pop()
         mcsimp = dern.split('=')
         d['mcsimp'] = '.*%s' % mcsimp.pop(0)
         if len(mcsimp) > 0:
            d['valeur'] = ' *=.*%s' % mcsimp.pop()
            assert(len(txt) == 0, '>>> trop de valeurs : %s' % dern)
      if len(txt) > 0:
         d['mcfact'] = '.*%s *= *_F *\(' % txt.pop()
      assert(len(txt) == 0, '>>> trop de valeurs : %s' % txt0)
      
      expr = '%(cmde)s *\(%(mcfact)s%(mcsimp)s%(valeur)s' % d
      return expr

# -----------------------------------------------------------------------------
   def add_filter(self, typ, *args):
      """Ajoute un ou plusieurs filtres.
         typ = 'regexp', 'para' ou 'user'.
      """
      if typ == 'command':
         assert len(args) == 1
         expr = self.decode_cmde(args[0])
         if expr is not None:
            self.filter.append(FILTER_REGEXP(expr))
      elif typ == 'regexp':
         assert len(args) == 1
         expr = args[0]
         if expr is not None:
            self.filter.append(FILTER_REGEXP(expr))
      elif typ == 'para':
         assert len(args) == 1
         expr = self.decode_critere(args[0])
         if expr is not None:
            self.filter.append(FILTER_PARA(expr))
      elif typ == 'user':
         assert len(args) == 1
         d = globals().copy()
         execfile(args[0], d)
         if d.has_key('user_filter'):
            if type(d['user_filter']) in (list, tuple):
               l_filter = d['user_filter']
            else:
               l_filter = [d['user_filter'],]
            for ff in l_filter:
               assert isinstance(ff, FILTER)
            self.filter.extend(l_filter)
      else:
         print '>>> unknown filter type (%s) : ignored.' % typ

# -----------------------------------------------------------------------------
   def build_list(self):
      """Applique les filtres et retourne la liste des cas-test obtenue.
      """
      l_type = [type(f) for f in self.filter]
      need_content = FILTER_REGEXP in l_type
      need_para    = FILTER_PARA in l_type
      ignored = Set()
      if self.verbose:
         p = Progress(maxi=len(self.liste_ct), format='%5.1f %%',
                     msg='Analyse des cas-tests... ')
      n_ign = 0
      for i, ct in enumerate(self.liste_ct):
         if self.verbose:
            p.Update(i)
         # texte du jdc
         if need_content:
            l_content = []
            for fic in glob(self.filename(ct, 'com*')):
               l_content.append(open(fic, 'r').read())
            jdc = os.linesep.join(l_content)
         else:
            jdc = ''
         # paramtres du test
         if need_para:
            iret, d_para, err = self.read_para_ct(ct)
            if iret != 0:
               n_ign += 1
               if self.verbose:
                  print err
         else:
            d_para = {}
         # application des filtres
         for filter in self.filter:
            if not filter.check(jdc=jdc, para=d_para, test=self.filename(ct, '')):
               ignored.add(ct)
               break
      if self.verbose:
         p.End()
      res = list(Set(self.liste_ct).difference(ignored))
      res.sort()
      if self.verbose:
         print '%6d tests analyss sur %d (%d ignors)' \
            % (len(self.liste_ct) - n_ign, len(self.liste_ct), n_ign)
         print '%6d tests dans la liste' % len(res)
      return res

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def TestList(run, *args):
   """Build a list of testcases.
   """
   REPREF = os.path.join(run.aster_root, run['aster_vers'])
   if not run.get('config'):
      fconf = os.path.join(REPREF, 'config.txt')
   else:
      fconf = os.path.abspath(run['config'])
   conf = ASTER_CONFIG(fconf, run)

   l_ct = list(args)
   if len(l_ct) < 1 and not (run['all_test'] or run.get('test_list')):
      run.parser.error(
            _("'--%s' : you must give testcase names or use one of --all/--test_list option") \
            % run.current_action)
   if run.get('test_list'):
      iret, l_lue = get_list(run['test_list'])
      if iret == 0:
         l_ct.extend(l_lue)
      else:
         run.Mess(_('error during reading file : %s') % flist, _('<F>_ERROR'))
   
   astest_dir = run.get('astest_dir', os.path.join(REPREF, conf['SRCTEST'][0]))
   
   run.PrintExitCode = False
   if run['debug']:
      run.DBG('astest_dir', astest_dir)
      run.DBG('filter', run.get('filter'))
      run.DBG('command', run.get('command'))
      run.DBG('search', run.get('search'))
   
   filtre = LISTE_CT(astest_dir=astest_dir, all=run['all_test'], verbose=not run['silent'])
   filtre.add_ct(l_ct)
   # add filters on parameters
   if run.get('filter'):
      for expr in run['filter']:
         filtre.add_filter('para', expr)
   # add filters based on command used
   for expr in run.get('command', []):
      filtre.add_filter('command', expr)
   # add filters based on regular expression
   for expr in run.get('search', []):
      filtre.add_filter('regexp', expr)
   # add user filters
   if run.get('user_filter'):
      for f in run['user_filter']:
         filtre.add_filter('user', f)
      
   res = filtre.build_list()
   
   result = os.linesep.join(res)
   if run.get('output'):
      open(run['output'], 'w').write(result)
      print 'Liste dans le fichier : %s' % run['output']
   else:
      print result

