# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""Definition of MPI_INFO class.

Synopsis:
      as_run --serv fich.export
               V     directement (interactif)
                  ou via sub_script (batch)
      as_run fich.export
               V
      as_exec.py
               V
      PrepEnv dans /shared_tmp/interactif.XXX
               V
      mpirun script.sh -wd /shared_tmp/interactif.XXX
               >  parallel_cp tracker /shared_tmp/interactif.XXX /local_tmp/interactif.XXX.proc.#i
               >  cd /local_tmp/interactif.XXX.proc.#i
               >  ./asteru_mpi args
               >  if #i == 0:
                     cp /local_tmp/interactif.XXX.proc.#i/results /shared_tmp/interactif.XXX
"""

__revision__ = "$Id: as_mpi.py 3750 2009-01-15 15:50:06Z courtois $"


# ----- differ messages translation
def _(mesg):
   return mesg

import os
import re

#-------------------------------------------------------------------------------
class MPI_INFO:
   """Store informations for MPI executions.
   """
   def __init__(self, defines=None):
      """Initialization needs "DEFS" field of ASTER_CONFIG object.
      """
      if defines is not None:
         self._use_mpi = '_USE_MPI' in defines
      else:
         self._use_mpi = False
      self._nbnode = 1
      self._nbcpu  = 1
      self.global_reptrav = ''
      self.local_reptrav  = ''
      
      self.timer_comment = _("""(*) cpu and system times are not correctly counted using mpirun.""")
   
#-------------------------------------------------------------------------------
   def set_cpuinfo(self, nbnode, nbcpu):
      """Set values of number of processors, of nodes.
      """
      try:
         self._nbcpu = int(nbcpu)
      except ValueError:
         self._nbcpu = 1
      
      try:
         self._nbnode = int(nbnode)
      except ValueError:
         self._nbnode = 1
      
      self._nbnode = min(self._nbnode, self._nbcpu)
      if not self._use_mpi:
         self._nbnode = 1
         self._nbcpu  = 1

#-------------------------------------------------------------------------------
   def really(self):
      """Returns True if Code_Aster executions need mpirun.
      """
      return self._use_mpi

#-------------------------------------------------------------------------------
   def set_rep_trav(self, run, reptrav, basename=''):
      """Set temporary directory for Code_Aster executions.
      """
      if reptrav:
         self.local_reptrav = self.global_reptrav = reptrav
      else:
         dirname = run['rep_trav']
         self.local_reptrav = os.path.join(dirname, basename)
         # used shared directory only if more than one node
         if self.really() and self.nbnode() > 1:
            dirname = run['shared_tmp']
         self.global_reptrav = os.path.join(dirname, basename)

#-------------------------------------------------------------------------------
   def get_exec_command(self, run, cmd_in, add_tee=False):
      """Returns command to run Code_Aster.
      """
      if not self.really():
         if add_tee:
            cmd_in = """( %(cmd_in)s ; echo %(var)s=$? ) | tee fort.6 ; exit `grep -a %(var)s fort.6 | sed -e 's/%(var)s=//'`""" % {
               'cmd_in' : cmd_in,
               'var'    : 'EXECUTION_CODE_ASTER_EXIT_%s' % run['num_job'],
            }
         return cmd_in

      mpi_script = os.path.join(self.global_reptrav, 'mpi_script.sh')
      
      if self.really() and self.nbnode() > 1:
         cp_cmd = '%s --with-as_run --remote_copy_protocol=%s --remote_shell_protocol=%s %s' \
            % (os.path.join(run.astk_serv_root, 'bin', 'parallel_cp'),
               run['remote_copy_protocol'], run['remote_shell_protocol'],
               self.global_reptrav
#               os.path.join(run['shared_tmp'], 'tracker.%s' % run['num_job'])
               )
      else:
         cp_cmd = 'cp -r'
      
      dict_mpi_args = {
         'mpirun_cmd'         : run['mpirun_cmd'],
         'mpi_get_procid_cmd' : run['mpi_get_procid_cmd'],
         'mpi_hostfile'       : run['mpi_hostfile'],
         'mpi_nbnoeud'        : self.nbnode(),
         'mpi_nbcpu'          : self.nbcpu(),
         'cmd_to_run'         : cmd_in,
         'wrkdir'             : self.global_reptrav,
         'local_wrkdir'       : self.local_reptrav + '.',
         'program'            : mpi_script,
         'cp_cmd'             : cp_cmd,
         'num_job'            : run['num_job'],
      }
      template = os.path.join(run.astk_serv_root, 'bin', 'mpirun_template')
      if not run.Exists(template):
         run.Mess(_('file not found : %s'), _('<F>_FILE_NOT_FOUND'))
      content = open(template, 'r').read() % dict_mpi_args
      run.DBG(content, all=True)
      open(mpi_script, 'w').write(content)
      os.chmod(mpi_script, 0755)
      return run['mpirun_cmd'] % dict_mpi_args

#-------------------------------------------------------------------------------
   def nbcpu(self):
      return self._nbcpu

#-------------------------------------------------------------------------------
   def nbnode(self):
      return self._nbnode

#-------------------------------------------------------------------------------
   def reptrav(self, procid=None):
      """Returns global workdir or local workdir on proc #i.
      """
      if procid is None or not self.really():
         res = self.global_reptrav
      else:
         res = '%s.proc.%d' % (self.local_reptrav, procid)
      return res

#-------------------------------------------------------------------------------
   def add_to_timer(self, run, text, title):
      """Add info cpus to the timer reading 'text'.
      """
      if not hasattr(run, 'timer') or not self.really():
         return
      l_info_cpu = re.findall('PROC=([0-9]+) INFO_CPU=(.*)', text)
      l_info_cpu.sort()
      for procid, infos in l_info_cpu:
         cpu_mpi, sys_mpi = 0., 0.
         id_timer = '  %s - proc #%s' % (title, procid)
         run.timer.Stop(id_timer)
         try:
            l_ch = re.sub(' +', ' ', infos.strip()).split(' ')
            assert len(l_ch) == 4, 'cpu+sys, cpu, sys, elapsed'
            cpu_mpi = float(l_ch[1])
            sys_mpi = float(l_ch[2])
         except (AssertionError, TypeError):
            run.Mess(_('Unable to retreive CPU info for proc #%s') % procid)
            run.DBG('Info CPU proc #%s' % procid, infos, l_ch)
         run.timer.Add(id_timer, cpu_mpi, sys_mpi)
         # add time spend in 'Code_Aster run' and to total
         run.timer.Add(title, cpu_mpi, sys_mpi, to_total=True)

#-------------------------------------------------------------------------------
   def add_timer_comment(self, run):
      """Add a warning when mpirun is called without Aster.
      """
      if hasattr(run, 'timer') and self.really():
         run.timer.AddComment(self.timer_comment, init=True)

