# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""
   Dfinit des objets permettant la gestion/rpartition de calculs
   sur plusieurs machines.
"""

__revision__ = "$Id: as_repart.py 3684 2008-12-15 08:01:04Z courtois $"


# ----- differ messages translation
def _(mesg):
   return mesg

import os
import time
import traceback
from pprint import pprint, pformat

from as_thread import Lock
from utils import parse_config

_DBG_tri   = False
_DBG_crit  = False
_DBG_alloc = False

# ------------------------------------------------------------------------------
def isnum(val):
   return type(val) in (int, long, float)

# ------------------------------------------------------------------------------
# constants
NORESOURCE =  1
ALLOCATED  =  0
OVERLIMIT  = -1

# ------------------------------------------------------------------------------
class ResourceManagerError(Exception):
   pass

# ------------------------------------------------------------------------------
class ResourceManager:
   """Class to manage resources to run a lot of calculations of several hosts.
   """
   lock = Lock()

   def __init__(self, hostinfo):
      """Initializations."""
      self.hostinfo = hostinfo.copy()
      self.all_hosts = self.hostinfo.keys()
      
      self.d_crit = {# numkey,   default, reverse order
         # criteria
         'cpurun' : ('000_cpurun',  0,     False, ),
         'memrun' : ('010_memrun',  0,     False, ),
         # characteristics of hosts (constant)
         'mem'    : ('101_mem',     0,     True,  ),
         'cpu'    : ('102_cpu',     0,     True,  ),
         'nomjob' : ('998_job',    '',     False, ),
         'host'   : ('999_host',   '',     False, ),
      }
      self.l_crit = [(v[0], v[2]) for v in self.d_crit.values()]
      self.l_crit.sort()
      self.job_keys = [k for k, v in self.d_crit.items() if v[0] > '100_']
      # build the list of all infos
      self.infos = []
      self.limit = {}
      for host, info in self.hostinfo.items():
         info['host'] = host
         d = {}
         for crit, val in self.d_crit.items():
            numkey, default, reverse = val
            d[numkey] = info.get(crit, default)
            if numkey > '100_' and numkey < '900_' \
            and ( \
               self.limit.get(crit) is None or (reverse and self.limit[crit] < d[numkey]) \
               or  (not reverse and self.limit[crit] > d[numkey]) \
            ):
               self.limit[crit] = d[numkey]
         self.infos.append(d)
      # to store job parameters
      self.history = {}

# ------------------------------------------------------------------------------
   def get(self, host, key, default=None):
      """Return a current value."""
      res = default
      if self.d_crit[key][0] > '100_':
         res = self.hostinfo[host][key]
      else:
         nkey = self.d_crit[key][0]
         for info in self.infos:
            if info['999_host'] == host:
               res = info[nkey]
               break
      return res

# ------------------------------------------------------------------------------
   def set(self, host, key, value):
      """Set a current value."""
      done = False
      if self.d_crit[key][0] > '100_':
         raise ResourceManagerError, "can not be changed : '%s'" % key
      else:
         nkey = self.d_crit[key][0]
         for info in self.infos:
            if info['999_host'] == host:
               info[nkey] = value
               done = True
      if not done:
         raise ResourceManagerError, "can't set '%s'" % key

# ------------------------------------------------------------------------------
   def add(self, host, key, value):
      """Add 'value' to the current value."""
      current = self.get(host, key)
      self.set(host, key, current + value)

# ------------------------------------------------------------------------------
   def sub(self, host, key, value):
      """Substract 'value' to the current value."""
      current = self.get(host, key)
      self.set(host, key, current - value)

# ------------------------------------------------------------------------------
   def store_job(self, host='unknown', **kwjob):
      """Store 'kwjob' in history."""
      if host is None:
         return
      dico = kwjob.copy()
      dico['host'] = host
      dico['allocated'] = time.strftime('%a %H:%M:%S')
      self.history[kwjob['nomjob']] = dico

# ------------------------------------------------------------------------------
   def get_job(self, jobname):
      """Get 'kwjob' from history."""
      dico = self.history.get(jobname)
      if dico is None:
         return {}
      dico['released'] = time.strftime('%a %H:%M:%S')
      return dico.copy()

# ------------------------------------------------------------------------------
   def get_history(self):
      """Return a copy of the jobs's 'history'."""
      dico = self.history.copy()
      return dico

# ------------------------------------------------------------------------------
   def action(self, what, *args, **kwargs):
      """Run safely a method which access to infos attribute."""
      result = None
      self.lock.acquire()
      tberr = None
      try:
         result = getattr(self, what)(*args, **kwargs)
      except Exception, err:
         tberr = traceback.format_exc()
      self.lock.release()
      if tberr:
         raise ResourceManagerError, tberr
      return result

# ------------------------------------------------------------------------------
   def get_first(self, values=None):
      """Return the most available host."""
      if values is None:
         values = self.infos[:]
      if len(values) == 0:
         return {}
      for crit, rev in self.l_crit:
         values.sort(reverse=rev)
         if _DBG_tri:
            print crit
            pprint(values)
         val0 = values[0][crit]
         new = [values[0],]
         for info in values[1:]:
            if info[crit] != val0:
               break
            new.append(info)
         values = new
         if len(values) == 1:
            break
      if _DBG_tri:
         print '--- FIN ---'
         pprint(values)
      return values[0]

# ------------------------------------------------------------------------------
   def suitable_host(self, **kwjob):
      """Limit infos to capable hosts."""
      values = []
      for info in self.infos:
         isok = True
         for par, val in kwjob.items():
            if not self.isok(par, val, info):
               isok = False
               break
         if isok:
            values.append(info)
      return values

# ------------------------------------------------------------------------------
   def available_host(self, **kwjob):
      """Return the most available host accepting job parameters."""
      if _DBG_crit:
         print 'job parameters : ',
         pprint(kwjob)
      values = self.suitable_host(**kwjob)
      if _DBG_crit:
         print '%d suitable hosts : %s' \
            % (len(values), [info[self.d_crit['host'][0]] for info in values])
      avail = self.get_first(values)
      return avail

# ------------------------------------------------------------------------------
   def isok(self, key, value, info):
      """Tell if 'value' is under the limit vs 'info[par]'.
      """
      if not key in self.job_keys:
         return False
      val_ref = info[self.d_crit[key][0]]
      if not self.d_crit.has_key(key + 'run'):
         return True
      val_run = info[self.d_crit[key + 'run'][0]]
      ok = (value + val_run) <= val_ref
      if _DBG_crit:
         rep = '-'
         if ok: rep = 'ok'
         print '%-2s host=%-24s para=%-4s allocated=%-4s requested=%-4s ref=%-4s' % \
            (rep, info[self.d_crit['host'][0]], key, val_run, value, val_ref)
      return ok

# ------------------------------------------------------------------------------
   def Request(self, **kwjob):
      """Ask for an available host and block resources."""
      info = self.action('available_host', **kwjob)
      host = info.get(self.d_crit['host'][0], None)
      if _DBG_alloc:
         print 'job allocated on %s : %s' % (host, pformat(kwjob))
      if host is not None:
         status = ALLOCATED
         for key in self.job_keys:
            if isnum(kwjob.get(key)):
               self.action('add', host, key + 'run', kwjob[key])
      else:
         status = NORESOURCE
         for key, lim in self.limit.items():
            if kwjob.get(key) and (\
               (self.d_crit[key][2]     and kwjob[key] > lim) or \
               (not self.d_crit[key][2] and kwjob[key] < lim) ):
               status = OVERLIMIT
               break
      self.action('store_job', host, **kwjob)
      if _DBG_alloc:
         print self.Load()
      return host, status

# ------------------------------------------------------------------------------
   def Free(self, jobname):
      """Free job resources on 'host'."""
      kwjob = self.action('get_job', jobname)
      if not kwjob:
         return
      host = kwjob['host']
      ddbg = {}
      for key in self.job_keys:
         if isnum(kwjob.get(key)):
            self.action('sub', host, key + 'run', kwjob[key])
            ddbg[key] = kwjob[key]
      if _DBG_alloc:
         print 'job released : %s' % (pformat(ddbg))

# ------------------------------------------------------------------------------
   def repr_history(self):
      """Returns jobs's history."""
      dico = self.action('get_history')
      ljob = [(v['allocated'], k) for k,v in dico.items()]
      ljob.sort()
      head= '    job      host      allocated    cpu    mem     released'
      fmt = '%(job_)-14s %(host_)-16s %(allocated)12s %(cpu)6d %(mem)6d %(released)12s'
      dini = { 'job_' : '', 'host_' : '', 'allocated' : '', 'cpu' : 0, 'mem' : 0, 'released' : '' }
      txt = [head,]
      for start, job in ljob:
         d = dini.copy()
         d.update(dico[job])
         d.update({ 'job_' : job[:14], 'host_' : d['host'][:16] })
         txt.append(fmt % d)
      return os.linesep.join(txt)

# ------------------------------------------------------------------------------
   def Load(self):
      """Returns current load."""
      infos = self.infos[:]
      lkey = ['cpu', 'mem']
      head= '    host      cpu    mem'
      fmt = '%(host_)-16s %(cpu)6d %(mem)6d'
      txt = [head,]
      for info in infos:
         d = { 'host_' : info[self.d_crit['host'][0]][:16] }
         for k in lkey:
            d[k] = info[self.d_crit[k + 'run'][0]]
         txt.append(fmt % d)
      return os.linesep.join(txt)

# ------------------------------------------------------------------------------
   def GetConfig(self, host):
      """Returns the configuration of 'host'."""
      return self.hostinfo.get(host, {}).copy()

# ------------------------------------------------------------------------------
   def get_sum(self, key):
      """Returns the sum of a resource."""
      return sum([self.get(host, key, 0) for host in self.all_hosts])


# ------------------------------------------------------------------------------
class PureLocalResource(ResourceManager):
   """Derived class to run only on local host.
   """
   def __init__(self, run):
      """Initialization using ASTER_RUN object."""
      host = run.GetHostName().split('.')[0]
      cpu  = run.GetCpuInfo('numcpu')
      mem  = run.GetMemInfo('memtotal') or 99e3
      c = { host : { 'host' : host, 'cpu' : cpu, 'mem' : mem }}
      ResourceManager.__init__(self, hostinfo=c)


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
if __name__ == '__main__':
   # unittest
   if False:
      from as_run import AsRunFactory
      run=AsRunFactory('/usr/local/aster/ASTK/ASTK_SERV')
      rc = PureLocalResource(run)
   
   
   else:
      from utils import parse_config
      c = parse_config('../unittest/astout/hostfile')
      #for v in c.values():
      #   del v['mem']
      rc = ResourceManager(c)
   
   if False:
      print 'cpu =', rc.action('get', 'cli70cx', 'cpu')
      print 'mem =', rc.action('get', 'cli70cx', 'mem')
      print 'mem used =', rc.action('get', 'cli70cx', 'memrun')
      
      assert rc.isok('mem', 384, rc.infos[0]) is True
      rc.action('add', 'cli70cx', 'memrun', 256)
      print 'mem used =', rc.action('get', 'cli70cx', 'memrun')
      assert rc.isok('mem', 384, rc.infos[0]) is False
      rc.action('sub', 'cli70cx', 'memrun', 256)
      print 'mem used =', rc.action('get', 'cli70cx', 'memrun')

   host1, status = rc.Request(nomjob='job1', cpu=1, mem=256)
   #time.sleep(1)
   host, status = rc.Request(cpu=1, mem=256, time=44)
   assert host is None and status == NORESOURCE
   host, status = rc.Request(cpu=1, mem=2048, time=14)
   assert host is None and status == OVERLIMIT
   host2, status = rc.Request(nomjob='job2', cpu=1, mem=128)
   #time.sleep(1)
   print
   rc.Free('job2')
   print
   host3, status = rc.Request(nomjob='job3', cpu=2, mem=64)
   #time.sleep(1)

   rc.Free('job1')
   #time.sleep(1)
   host4, status = rc.Request(nomjob='job4', cpu=1)
   rc.Free('job3')
   rc.Free('job4')

   print rc.repr_history()

